@@ -1,5 +1,5 @@
 /*
-Copyright (C) 2012-2015 Ben Kurtovic <ben.kurtovic@gmail.com>
+Copyright (C) 2012-2018 Ben Kurtovic <ben.kurtovic@gmail.com>
 
 Permission is hereby granted, free of charge, to any person obtaining a copy of
 this software and associated documentation files (the "Software"), to deal in
@@ -22,6 +22,7 @@ SOFTWARE.
 
 #include "tok_parse.h"
 #include "contexts.h"
+#include "definitions.h"
 #include "tag_data.h"
 #include "tok_support.h"
 #include "tokens.h"
@@ -29,17 +30,11 @@ SOFTWARE.
 #define DIGITS    "0123456789"
 #define HEXDIGITS "0123456789abcdefABCDEF"
 #define ALPHANUM  "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
+#define URISCHEME "abcdefghijklmnopqrstuvwxyz0123456789+.-"
 
 #define MAX_BRACES 255
 #define MAX_ENTITY_SIZE 8
 
-#define GET_HTML_TAG(markup) (markup == ':' ? "dd" : markup == ';' ? "dt" : "li")
-#define IS_PARSABLE(tag) (call_def_func("is_parsable", tag, NULL))
-#define IS_SINGLE(tag) (call_def_func("is_single", tag, NULL))
-#define IS_SINGLE_ONLY(tag) (call_def_func("is_single_only", tag, NULL))
-#define IS_SCHEME(scheme, slashes) \
-    (call_def_func("is_scheme", scheme, slashes ? Py_True : Py_False))
-
 typedef struct {
     PyObject* title;
     int level;
@@ -47,6 +42,8 @@ typedef struct {
 
 /* Forward declarations */
 
+static PyObject* Tokenizer_really_parse_external_link(
+    Tokenizer*, int, Textbuffer*);
 static int Tokenizer_parse_entity(Tokenizer*);
 static int Tokenizer_parse_comment(Tokenizer*);
 static int Tokenizer_handle_dl_term(Tokenizer*);
@@ -79,21 +76,6 @@ static int heading_level_from_context(uint64_t n)
     return level;
 }
 
-/*
-    Call the given function in definitions.py, using 'in1' and 'in2' as
-    parameters, and return its output as a bool.
-*/
-static int call_def_func(const char* funcname, PyObject* in1, PyObject* in2)
-{
-    PyObject* func = PyObject_GetAttrString(definitions, funcname);
-    PyObject* result = PyObject_CallFunctionObjArgs(func, in1, in2, NULL);
-    int ans = (result == Py_True) ? 1 : 0;
-
-    Py_DECREF(func);
-    Py_DECREF(result);
-    return ans;
-}
-
 /*
     Sanitize the name of a tag so it can be compared with others for equality.
 */
@@ -267,7 +249,7 @@ static int Tokenizer_handle_template_param(Tokenizer* self)
     else if (self->topstack->context & LC_TEMPLATE_PARAM_VALUE)
         self->topstack->context ^= LC_TEMPLATE_PARAM_VALUE;
     if (self->topstack->context & LC_TEMPLATE_PARAM_KEY) {
-        stack = Tokenizer_pop_keeping_context(self);
+        stack = Tokenizer_pop(self);
         if (!stack)
             return -1;
         if (Tokenizer_emit_all(self, stack)) {
@@ -292,7 +274,7 @@ static int Tokenizer_handle_template_param_value(Tokenizer* self)
 {
     PyObject *stack;
 
-    stack = Tokenizer_pop_keeping_context(self);
+    stack = Tokenizer_pop(self);
     if (!stack)
         return -1;
     if (Tokenizer_emit_all(self, stack)) {
@@ -319,7 +301,7 @@ static PyObject* Tokenizer_handle_template_end(Tokenizer* self)
             return Tokenizer_fail_route(self);
     }
     else if (self->topstack->context & LC_TEMPLATE_PARAM_KEY) {
-        stack = Tokenizer_pop_keeping_context(self);
+        stack = Tokenizer_pop(self);
         if (!stack)
             return NULL;
         if (Tokenizer_emit_all(self, stack)) {
@@ -362,30 +344,70 @@ static PyObject* Tokenizer_handle_argument_end(Tokenizer* self)
 static int Tokenizer_parse_wikilink(Tokenizer* self)
 {
     Py_ssize_t reset;
-    PyObject *wikilink;
+    PyObject *extlink, *wikilink, *kwargs;
 
+    reset = self->head + 1;
     self->head += 2;
-    reset = self->head - 1;
-    wikilink = Tokenizer_parse(self, LC_WIKILINK_TITLE, 1);
+    // If the wikilink looks like an external link, parse it as such:
+    extlink = Tokenizer_really_parse_external_link(self, 1, NULL);
     if (BAD_ROUTE) {
         RESET_ROUTE();
+        self->head = reset + 1;
+        // Otherwise, actually parse it as a wikilink:
+        wikilink = Tokenizer_parse(self, LC_WIKILINK_TITLE, 1);
+        if (BAD_ROUTE) {
+            RESET_ROUTE();
+            self->head = reset;
+            if (Tokenizer_emit_text(self, "[["))
+                return -1;
+            return 0;
+        }
+        if (!wikilink)
+            return -1;
+        if (Tokenizer_emit(self, WikilinkOpen)) {
+            Py_DECREF(wikilink);
+            return -1;
+        }
+        if (Tokenizer_emit_all(self, wikilink)) {
+            Py_DECREF(wikilink);
+            return -1;
+        }
+        Py_DECREF(wikilink);
+        if (Tokenizer_emit(self, WikilinkClose))
+            return -1;
+        return 0;
+    }
+    if (!extlink)
+        return -1;
+    if (self->topstack->context & LC_EXT_LINK_TITLE) {
+        // In this exceptional case, an external link that looks like a
+        // wikilink inside of an external link is parsed as text:
+        Py_DECREF(extlink);
         self->head = reset;
         if (Tokenizer_emit_text(self, "[["))
             return -1;
         return 0;
     }
-    if (!wikilink)
-        return -1;
-    if (Tokenizer_emit(self, WikilinkOpen)) {
-        Py_DECREF(wikilink);
+    if (Tokenizer_emit_text(self, "[")) {
+        Py_DECREF(extlink);
         return -1;
     }
-    if (Tokenizer_emit_all(self, wikilink)) {
-        Py_DECREF(wikilink);
+    kwargs = PyDict_New();
+    if (!kwargs) {
+        Py_DECREF(extlink);
         return -1;
     }
-    Py_DECREF(wikilink);
-    if (Tokenizer_emit(self, WikilinkClose))
+    PyDict_SetItemString(kwargs, "brackets", Py_True);
+    if (Tokenizer_emit_kwargs(self, ExternalLinkOpen, kwargs)) {
+        Py_DECREF(extlink);
+        return -1;
+    }
+    if (Tokenizer_emit_all(self, extlink)) {
+        Py_DECREF(extlink);
+        return -1;
+    }
+    Py_DECREF(extlink);
+    if (Tokenizer_emit(self, ExternalLinkClose))
         return -1;
     return 0;
 }
@@ -417,12 +439,14 @@ static PyObject* Tokenizer_handle_wikilink_end(Tokenizer* self)
 */
 static int Tokenizer_parse_bracketed_uri_scheme(Tokenizer* self)
 {
-    static const char* valid = "abcdefghijklmnopqrstuvwxyz0123456789+.-";
+    static const char* valid = URISCHEME;
     Textbuffer* buffer;
     PyObject* scheme;
     Unicode this;
     int slashes, i;
 
+    if (Tokenizer_check_route(self, LC_EXT_LINK_URI) < 0)
+        return 0;
     if (Tokenizer_push(self, LC_EXT_LINK_URI))
         return -1;
     if (Tokenizer_read(self, 0) == '/' && Tokenizer_read(self, 1) == '/') {
@@ -439,7 +463,7 @@ static int Tokenizer_parse_bracketed_uri_scheme(Tokenizer* self)
             while (1) {
                 if (!valid[i])
                     goto end_of_loop;
-                if (this == valid[i])
+                if (this == (Unicode) valid[i])
                     break;
                 i++;
             }
@@ -474,7 +498,7 @@ static int Tokenizer_parse_bracketed_uri_scheme(Tokenizer* self)
         Textbuffer_dealloc(buffer);
         if (!scheme)
             return -1;
-        if (!IS_SCHEME(scheme, slashes)) {
+        if (!is_scheme(scheme, slashes)) {
             Py_DECREF(scheme);
             Tokenizer_fail_route(self);
             return 0;
@@ -489,12 +513,13 @@ static int Tokenizer_parse_bracketed_uri_scheme(Tokenizer* self)
 */
 static int Tokenizer_parse_free_uri_scheme(Tokenizer* self)
 {
-    static const char* valid = "abcdefghijklmnopqrstuvwxyz0123456789+.-";
+    static const char* valid = URISCHEME;
     Textbuffer *scheme_buffer = Textbuffer_new(&self->text);
     PyObject *scheme;
     Unicode chunk;
     Py_ssize_t i;
     int slashes, j;
+    uint64_t new_context;
 
     if (!scheme_buffer)
         return -1;
@@ -511,7 +536,7 @@ static int Tokenizer_parse_free_uri_scheme(Tokenizer* self)
                 FAIL_ROUTE(0);
                 return 0;
             }
-        } while (chunk != valid[j++]);
+        } while (chunk != (Unicode) valid[j++]);
         Textbuffer_write(scheme_buffer, chunk);
     }
     end_of_loop:
@@ -523,14 +548,19 @@ static int Tokenizer_parse_free_uri_scheme(Tokenizer* self)
     }
     slashes = (Tokenizer_read(self, 0) == '/' &&
                Tokenizer_read(self, 1) == '/');
-    if (!IS_SCHEME(scheme, slashes)) {
+    if (!is_scheme(scheme, slashes)) {
         Py_DECREF(scheme);
         Textbuffer_dealloc(scheme_buffer);
         FAIL_ROUTE(0);
         return 0;
     }
     Py_DECREF(scheme);
-    if (Tokenizer_push(self, self->topstack->context | LC_EXT_LINK_URI)) {
+    new_context = self->topstack->context | LC_EXT_LINK_URI;
+    if (Tokenizer_check_route(self, new_context) < 0) {
+        Textbuffer_dealloc(scheme_buffer);
+        return 0;
+    }
+    if (Tokenizer_push(self, new_context)) {
         Textbuffer_dealloc(scheme_buffer);
         return -1;
     }
@@ -553,7 +583,7 @@ static int Tokenizer_handle_free_link_text(
     Tokenizer* self, int* parens, Textbuffer* tail, Unicode this)
 {
     #define PUSH_TAIL_BUFFER(tail, error)                            \
-        if (tail->length > 0) {                                      \
+        if (tail && tail->length > 0) {                              \
             if (Textbuffer_concat(self->topstack->textbuffer, tail)) \
                 return error;                                        \
             if (Textbuffer_reset(tail))                              \
@@ -692,7 +722,6 @@ Tokenizer_remove_uri_scheme_from_textbuffer(Tokenizer* self, PyObject* link)
 */
 static int Tokenizer_parse_external_link(Tokenizer* self, int brackets)
 {
-    #define INVALID_CONTEXT self->topstack->context & AGG_NO_EXT_LINKS
     #define NOT_A_LINK                                        \
         if (!brackets && self->topstack->context & LC_DLTERM) \
             return Tokenizer_handle_dl_term(self);            \
@@ -702,7 +731,8 @@ static int Tokenizer_parse_external_link(Tokenizer* self, int brackets)
     PyObject *link, *kwargs;
     Textbuffer *extra;
 
-    if (INVALID_CONTEXT || !(Tokenizer_CAN_RECURSE(self))) {
+    if (self->topstack->context & AGG_NO_EXT_LINKS ||
+            !(Tokenizer_CAN_RECURSE(self))) {
         NOT_A_LINK;
     }
     extra = Textbuffer_new(&self->text);
@@ -978,7 +1008,7 @@ static int Tokenizer_really_parse_entity(Tokenizer* self)
         while (1) {
             if (!valid[j])
                 FAIL_ROUTE_AND_EXIT()
-            if (this == valid[j])
+            if (this == (Unicode) valid[j])
                 break;
             j++;
         }
@@ -1043,11 +1073,14 @@ static int Tokenizer_parse_entity(Tokenizer* self)
     Py_ssize_t reset = self->head;
     PyObject *tokenlist;
 
-    if (Tokenizer_push(self, 0))
+    if (Tokenizer_check_route(self, LC_HTML_ENTITY) < 0)
+        goto on_bad_route;
+    if (Tokenizer_push(self, LC_HTML_ENTITY))
         return -1;
     if (Tokenizer_really_parse_entity(self))
         return -1;
     if (BAD_ROUTE) {
+        on_bad_route:
         RESET_ROUTE();
         self->head = reset;
         if (Tokenizer_emit_char(self, '&'))
@@ -1247,6 +1280,7 @@ static int Tokenizer_handle_tag_data(
     else if (data->context & TAG_NOTE_SPACE) {
         if (data->context & TAG_QUOTED) {
             data->context = TAG_ATTR_VALUE;
+            Tokenizer_memoize_bad_route(self);
             trash = Tokenizer_pop(self);
             Py_XDECREF(trash);
             self->head = data->reset - 1;  // Will be auto-incremented
@@ -1284,7 +1318,12 @@ static int Tokenizer_handle_tag_data(
                 data->context |= TAG_QUOTED;
                 data->quoter = chunk;
                 data->reset = self->head;
-                if (Tokenizer_push(self, self->topstack->context))
+                if (Tokenizer_check_route(self, self->topstack->context) < 0) {
+                    RESET_ROUTE();
+                    data->context = TAG_ATTR_VALUE;
+                    self->head--;
+                }
+                else if (Tokenizer_push(self, self->topstack->context))
                     return -1;
                 return 0;
             }
@@ -1515,6 +1554,14 @@ static PyObject* Tokenizer_handle_single_tag_end(Tokenizer* self)
             if (depth == 0)
                 break;
         }
+        is_instance = PyObject_IsInstance(token, TagCloseSelfclose);
+        if (is_instance == -1)
+            return NULL;
+        else if (is_instance == 1) {
+            depth--;
+            if (depth == 0)  // Should never happen
+                return NULL;
+        }
     }
     if (!token || depth > 0)
         return NULL;
@@ -1552,6 +1599,8 @@ static PyObject* Tokenizer_really_parse_tag(Tokenizer* self)
 
     if (!data)
         return NULL;
+    if (Tokenizer_check_route(self, LC_TAG_OPEN) < 0)
+        return NULL;
     if (Tokenizer_push(self, LC_TAG_OPEN)) {
         TagData_dealloc(data);
         return NULL;
@@ -1570,6 +1619,7 @@ static PyObject* Tokenizer_really_parse_tag(Tokenizer* self)
                 if (data->context & TAG_QUOTED) {
                     // Unclosed attribute quote: reset, don't die
                     data->context = TAG_ATTR_VALUE;
+                    Tokenizer_memoize_bad_route(self);
                     trash = Tokenizer_pop(self);
                     Py_XDECREF(trash);
                     self->head = data->reset;
@@ -1592,11 +1642,11 @@ static PyObject* Tokenizer_really_parse_tag(Tokenizer* self)
             text = PyObject_GetAttrString(token, "text");
             if (!text)
                 return NULL;
-            if (IS_SINGLE_ONLY(text)) {
+            if (is_single_only(text)) {
                 Py_DECREF(text);
                 return Tokenizer_handle_single_only_tag_end(self);
             }
-            if (IS_PARSABLE(text)) {
+            if (is_parsable(text)) {
                 Py_DECREF(text);
                 return Tokenizer_parse(self, 0, 0);
             }
@@ -1644,7 +1694,7 @@ static int Tokenizer_handle_invalid_tag_start(Tokenizer* self)
                 Textbuffer_dealloc(buf);
                 return -1;
             }
-            if (!IS_SINGLE_ONLY(name))
+            if (!is_single_only(name))
                 FAIL_ROUTE(0);
             Py_DECREF(name);
             break;
@@ -2108,7 +2158,7 @@ Tokenizer_emit_table_tag(Tokenizer* self, const char* open_open_markup,
 /*
     Handle style attributes for a table until an ending token.
 */
-static PyObject* Tokenizer_handle_table_style(Tokenizer* self, char end_token)
+static PyObject* Tokenizer_handle_table_style(Tokenizer* self, Unicode end_token)
 {
     TagData *data = TagData_new(&self->text);
     PyObject *padding, *trash;
@@ -2142,6 +2192,7 @@ static PyObject* Tokenizer_handle_table_style(Tokenizer* self, char end_token)
                 if (data->context & TAG_QUOTED) {
                     // Unclosed attribute quote: reset, don't die
                     data->context = TAG_ATTR_VALUE;
+                    Tokenizer_memoize_bad_route(self);
                     trash = Tokenizer_pop(self);
                     Py_XDECREF(trash);
                     self->head = data->reset;
@@ -2168,18 +2219,22 @@ static PyObject* Tokenizer_handle_table_style(Tokenizer* self, char end_token)
 */
 static int Tokenizer_parse_table(Tokenizer* self)
 {
-    Py_ssize_t reset = self->head + 1;
-    PyObject *style, *padding;
+    Py_ssize_t reset = self->head;
+    PyObject *style, *padding, *trash;
     PyObject *table = NULL;
+    StackIdent restore_point;
     self->head += 2;
 
-    if(Tokenizer_push(self, LC_TABLE_OPEN))
+    if (Tokenizer_check_route(self, LC_TABLE_OPEN) < 0)
+        goto on_bad_route;
+    if (Tokenizer_push(self, LC_TABLE_OPEN))
         return -1;
     padding = Tokenizer_handle_table_style(self, '\n');
     if (BAD_ROUTE) {
+        on_bad_route:
         RESET_ROUTE();
         self->head = reset;
-        if (Tokenizer_emit_text(self, "{|"))
+        if (Tokenizer_emit_char(self, '{'))
             return -1;
         return 0;
     }
@@ -2192,13 +2247,18 @@ static int Tokenizer_parse_table(Tokenizer* self)
     }
 
     self->head++;
+    restore_point = self->topstack->ident;
     table = Tokenizer_parse(self, LC_TABLE_OPEN, 1);
     if (BAD_ROUTE) {
         RESET_ROUTE();
         Py_DECREF(padding);
         Py_DECREF(style);
+        while (!Tokenizer_IS_CURRENT_STACK(self, restore_point)) {
+            trash = Tokenizer_pop(self);
+            Py_XDECREF(trash);
+        }
         self->head = reset;
-        if (Tokenizer_emit_text(self, "{|"))
+        if (Tokenizer_emit_char(self, '{'))
             return -1;
         return 0;
     }
@@ -2221,7 +2281,7 @@ static int Tokenizer_parse_table(Tokenizer* self)
 */
 static int Tokenizer_handle_table_row(Tokenizer* self)
 {
-    PyObject *padding, *style, *row, *trash;
+    PyObject *padding, *style, *row;
     self->head += 2;
 
     if (!Tokenizer_CAN_RECURSE(self)) {
@@ -2231,14 +2291,13 @@ static int Tokenizer_handle_table_row(Tokenizer* self)
         return 0;
     }
 
-    if(Tokenizer_push(self, LC_TABLE_OPEN | LC_TABLE_ROW_OPEN))
+    if (Tokenizer_check_route(self, LC_TABLE_OPEN | LC_TABLE_ROW_OPEN) < 0)
+        return 0;
+    if (Tokenizer_push(self, LC_TABLE_OPEN | LC_TABLE_ROW_OPEN))
         return -1;
     padding = Tokenizer_handle_table_style(self, '\n');
-    if (BAD_ROUTE) {
-        trash = Tokenizer_pop(self);
-        Py_XDECREF(trash);
+    if (BAD_ROUTE)
         return 0;
-    }
     if (!padding)
         return -1;
     style = Tokenizer_pop(self);
@@ -2297,8 +2356,8 @@ Tokenizer_handle_table_cell(Tokenizer* self, const char *markup,
     if (cell_context & LC_TABLE_CELL_STYLE) {
         Py_DECREF(cell);
         self->head = reset;
-        if(Tokenizer_push(self, LC_TABLE_OPEN | LC_TABLE_CELL_OPEN |
-                          line_context))
+        if (Tokenizer_push(self, LC_TABLE_OPEN | LC_TABLE_CELL_OPEN |
+                           line_context))
             return -1;
         padding = Tokenizer_handle_table_style(self, '|');
         if (!padding)
@@ -2386,7 +2445,7 @@ static PyObject* Tokenizer_handle_end(Tokenizer* self, uint64_t context)
             text = PyObject_GetAttrString(token, "text");
             if (!text)
                 return NULL;
-            single = IS_SINGLE(text);
+            single = is_single(text);
             Py_DECREF(text);
             if (single)
                 return Tokenizer_handle_single_tag_end(self);
@@ -2519,6 +2578,8 @@ PyObject* Tokenizer_parse(Tokenizer* self, uint64_t context, int push)
     PyObject* temp;
 
     if (push) {
+        if (Tokenizer_check_route(self, context) < 0)
+            return NULL;
         if (Tokenizer_push(self, context))
             return NULL;
     }
@@ -2667,10 +2728,8 @@ PyObject* Tokenizer_parse(Tokenizer* self, uint64_t context, int push)
                 if (Tokenizer_parse_table(self))
                     return NULL;
             }
-            else if (Tokenizer_emit_char(self, this) || Tokenizer_emit_char(self, next))
+            else if (Tokenizer_emit_char(self, this))
                 return NULL;
-            else
-                self->head++;
         }
         else if (this_context & LC_TABLE_OPEN) {
             if (this == '|' && next == '|' && this_context & LC_TABLE_TD_LINE) {
