@@ -1,6 +1,6 @@
 # -*- coding: utf-8  -*-
 #
-# Copyright (C) 2012-2015 Ben Kurtovic <ben.kurtovic@gmail.com>
+# Copyright (C) 2012-2016 Ben Kurtovic <ben.kurtovic@gmail.com>
 #
 # Permission is hereby granted, free of charge, to any person obtaining a copy
 # of this software and associated documentation files (the "Software"), to deal
@@ -103,11 +103,10 @@ class TestTag(TreeEqualityTestCase):
         node1 = Tag(wraptext("i"), wraptext("foobar"))
         node2 = Tag(wraptext("math"), wraptext("foobar"))
         node3 = Tag(wraptext("br"), self_closing=True)
-        for a in (True, False):
-            for b in (True, False):
-                self.assertEqual("foobar", node1.__strip__(a, b))
-                self.assertEqual(None, node2.__strip__(a, b))
-                self.assertEqual(None, node3.__strip__(a, b))
+
+        self.assertEqual("foobar", node1.__strip__())
+        self.assertEqual(None, node2.__strip__())
+        self.assertEqual(None, node3.__strip__())
 
     def test_showtree(self):
         """test Tag.__showtree__()"""
