@@ -1,6 +1,6 @@
 # -*- coding: utf-8  -*-
 #
-# Copyright (C) 2012-2015 Ben Kurtovic <ben.kurtovic@gmail.com>
+# Copyright (C) 2012-2017 Ben Kurtovic <ben.kurtovic@gmail.com>
 #
 # Permission is hereby granted, free of charge, to any person obtaining a copy
 # of this software and associated documentation files (the "Software"), to deal
@@ -21,6 +21,7 @@
 # SOFTWARE.
 
 from __future__ import unicode_literals
+from difflib import unified_diff
 
 try:
     import unittest2 as unittest
@@ -30,6 +31,8 @@ except ImportError:
 from mwparserfromhell.compat import str
 from mwparserfromhell.nodes import HTMLEntity, Template, Text
 from mwparserfromhell.nodes.extras import Parameter
+from mwparserfromhell import parse
+
 from ._test_tree_equality import TreeEqualityTestCase, wrap, wraptext
 
 pgens = lambda k, v: Parameter(wraptext(k), wraptext(v), showkey=True)
@@ -67,12 +70,19 @@ class TestTemplate(TreeEqualityTestCase):
     def test_strip(self):
         """test Template.__strip__()"""
         node1 = Template(wraptext("foobar"))
-        node2 = Template(wraptext("foo"),
-                         [pgenh("1", "bar"), pgens("abc", "def")])
-        for a in (True, False):
-            for b in (True, False):
-                self.assertEqual(None, node1.__strip__(a, b))
-                self.assertEqual(None, node2.__strip__(a, b))
+        node2 = Template(wraptext("foo"), [
+            pgenh("1", "bar"), pgens("foo", ""), pgens("abc", "def")])
+        node3 = Template(wraptext("foo"), [
+            pgenh("1", "foo"),
+            Parameter(wraptext("2"), wrap([Template(wraptext("hello"))]),
+                      showkey=False),
+            pgenh("3", "bar")])
+
+        self.assertEqual(None, node1.__strip__(keep_template_params=False))
+        self.assertEqual(None, node2.__strip__(keep_template_params=False))
+        self.assertEqual("", node1.__strip__(keep_template_params=True))
+        self.assertEqual("bar def", node2.__strip__(keep_template_params=True))
+        self.assertEqual("foo bar", node3.__strip__(keep_template_params=True))
 
     def test_showtree(self):
         """test Template.__showtree__()"""
@@ -216,6 +226,7 @@ class TestTemplate(TreeEqualityTestCase):
         node39 = Template(wraptext("a"), [pgenh("1", " b ")])
         node40 = Template(wraptext("a"), [pgenh("1", " b"), pgenh("2", " c")])
         node41 = Template(wraptext("a"), [pgens("1", " b"), pgens("2", " c")])
+        node42 = Template(wraptext("a"), [pgens("b", "  \n")])
 
         node1.add("e", "f", showkey=True)
         node2.add(2, "g", showkey=False)
@@ -261,6 +272,7 @@ class TestTemplate(TreeEqualityTestCase):
         node39.add("1", "c")
         node40.add("3", "d")
         node41.add("3", "d")
+        node42.add("b", "hello")
 
         self.assertEqual("{{a|b=c|d|e=f}}", node1)
         self.assertEqual("{{a|b=c|d|g}}", node2)
@@ -278,7 +290,7 @@ class TestTemplate(TreeEqualityTestCase):
         self.assertIsInstance(node12.params[1].value.get(1), HTMLEntity)
         self.assertEqual("{{a|\nb = c|\nd = e|\nf = g|\nh = i}}", node13)
         self.assertEqual("{{a\n|b =c\n|d = e|f =g\n|h = i\n|j =k\n}}", node14)
-        self.assertEqual("{{a|b  = c\n|\nd  = e|\nf  =g |h  =i}}", node15)
+        self.assertEqual("{{a|b  = c\n|\nd  = e|\nf  =g |\nh  = i}}", node15)
         self.assertEqual("{{a|\nb = c|\nd = e|\nf = g|h=i}}", node16)
         self.assertEqual("{{a|b|c}}", node17)
         self.assertEqual("{{a|b|3=c}}", node18)
@@ -308,6 +320,7 @@ class TestTemplate(TreeEqualityTestCase):
         self.assertEqual("{{a|c}}", node39)
         self.assertEqual("{{a| b| c|d}}", node40)
         self.assertEqual("{{a|1= b|2= c|3= d}}", node41)
+        self.assertEqual("{{a|b=hello  \n}}", node42)
 
     def test_remove(self):
         """test Template.remove()"""
@@ -429,5 +442,172 @@ class TestTemplate(TreeEqualityTestCase):
         self.assertEqual("{{foo|a=b|c=d|e=f|a=|a=b}}", node26)
         self.assertRaises(ValueError, node27.remove, node28.get(1))
 
+    def test_formatting(self):
+        """test realistic param manipulation with complex whitespace formatting
+        (assumes that parsing works correctly)"""
+        tests = [
+    # https://en.wikipedia.org/w/index.php?title=Lamar_County,_Georgia&oldid=792356004
+    ("""{{Infobox U.S. county
+| county = Lamar County
+| state = Georgia
+| seal =
+| founded = 1920
+| seat wl = Barnesville
+| largest city wl = Barnesville
+| area_total_sq_mi = 186
+| area_land_sq_mi = 184
+| area_water_sq_mi = 2.3
+| area percentage = 1.3%
+| census yr = 2010
+| pop = 18317
+| density_sq_mi = 100
+| time zone = Eastern
+| footnotes =
+| web = www.lamarcountyga.com
+| ex image = Lamar County Georgia Courthouse.jpg
+| ex image cap = Lamar County courthouse in Barnesville
+| district = 3rd
+| named for = [[Lucius Quintus Cincinnatus Lamar II]]
+}}""",
+    """@@ -11,4 +11,4 @@
+ | area percentage = 1.3%
+-| census yr = 2010
+-| pop = 18317
++| census estimate yr = 2016
++| pop = 12345<ref>example ref</ref>
+ | density_sq_mi = 100"""),
+
+    # https://en.wikipedia.org/w/index.php?title=Rockdale_County,_Georgia&oldid=792359760
+    ("""{{Infobox U.S. County|
+ county = Rockdale County |
+ state = Georgia |
+ seal =  |
+ founded = October 18, 1870 |
+ seat wl = Conyers |
+ largest city wl = Conyers |
+ area_total_sq_mi = 132 |
+ area_land_sq_mi = 130 |
+ area_water_sq_mi = 2.3 |
+ area percentage = 1.7% |
+ census yr = 2010|
+ pop = 85215 |
+ density_sq_mi = 657 |
+ web = www.rockdalecounty.org
+| ex image = Rockdale-county-courthouse.jpg
+| ex image cap = Rockdale County Courthouse in Conyers
+| district = 4th
+| time zone= Eastern
+}}""",
+    """@@ -11,4 +11,4 @@
+  area percentage = 1.7% |
+- census yr = 2010|
+- pop = 85215 |
++ census estimate yr = 2016 |
++ pop = 12345<ref>example ref</ref> |
+  density_sq_mi = 657 |"""),
+
+    # https://en.wikipedia.org/w/index.php?title=Spalding_County,_Georgia&oldid=792360413
+    ("""{{Infobox U.S. County|
+| county = Spalding County |
+| state = Georgia |
+| seal =  |
+| founded = 1851 |
+| seat wl = Griffin |
+| largest city wl = Griffin |
+| area_total_sq_mi = 200 |
+| area_land_sq_mi = 196 |
+| area_water_sq_mi = 3.1 |
+| area percentage = 1.6% |
+| census yr = 2010|
+| pop = 64073 |
+| density_sq_mi = 326 |
+| web = www.spaldingcounty.com |
+| named for = [[Thomas Spalding]]
+| ex image = Spalding County Courthouse (NE corner).JPG
+| ex image cap = Spalding County Courthouse in Griffin
+| district = 3rd
+| time zone = Eastern
+}}""",
+    """@@ -11,4 +11,4 @@
+ | area percentage = 1.6% |
+-| census yr = 2010|
+-| pop = 64073 |
++|
++| census estimate yr = 2016 | pop = 12345<ref>example ref</ref> |
+ | density_sq_mi = 326 |"""),
+
+    # https://en.wikipedia.org/w/index.php?title=Clinton_County,_Illinois&oldid=794694648
+    ("""{{Infobox U.S. county
+ |county  = Clinton County
+ |state = Illinois
+| ex image           = File:Clinton County Courthouse, Carlyle.jpg
+| ex image cap       = [[Clinton County Courthouse (Illinois)|Clinton County Courthouse]]
+ |seal =
+ |founded = 1824
+ |named for = [[DeWitt Clinton]]
+ |seat wl= Carlyle
+| largest city wl = Breese
+ |time zone=Central
+ |area_total_sq_mi = 503
+ |area_land_sq_mi = 474
+ |area_water_sq_mi = 29
+ |area percentage = 5.8%
+ |census yr = 2010
+ |pop = 37762
+ |density_sq_mi = 80
+ |web = www.clintonco.illinois.gov
+| district = 15th
+}}""",
+    """@@ -15,4 +15,4 @@
+  |area percentage = 5.8%
+- |census yr = 2010
+- |pop = 37762
++ |census estimate yr = 2016
++ |pop = 12345<ref>example ref</ref>
+  |density_sq_mi = 80"""),
+
+    # https://en.wikipedia.org/w/index.php?title=Winnebago_County,_Illinois&oldid=789193800
+    ("""{{Infobox U.S. county |
+ county  = Winnebago County |
+ state = Illinois |
+ seal = Winnebago County il seal.png |
+ named for = [[Winnebago (tribe)|Winnebago Tribe]] |
+ seat wl= Rockford |
+ largest city wl = Rockford|
+ area_total_sq_mi = 519 |
+ area_land_sq_mi = 513|
+ area_water_sq_mi = 5.9 |
+ area percentage = 1.1% |
+ census yr = 2010|
+ pop = 295266 |
+ density_sq_mi = 575
+| web = www.wincoil.us
+| founded year = 1836
+| founded date = January 16
+| time zone = Central
+| district = 16th
+| district2 = 17th
+}}""",
+    """@@ -11,4 +11,4 @@
+  area percentage = 1.1% |
+- census yr = 2010|
+- pop = 295266 |
++ census estimate yr = 2016|
++ pop = 12345<ref>example ref</ref> |
+  density_sq_mi = 575""")]
+
+        for (original, expected) in tests:
+            code = parse(original)
+            template = code.filter_templates()[0]
+            template.add("pop", "12345<ref>example ref</ref>")
+            template.add('census estimate yr', "2016", before="pop")
+            template.remove("census yr")
+
+            oldlines = original.splitlines(True)
+            newlines = str(code).splitlines(True)
+            difflines = unified_diff(oldlines, newlines, n=1)
+            diff = "".join(list(difflines)[2:]).strip()
+            self.assertEqual(expected, diff)
+
 if __name__ == "__main__":
     unittest.main(verbosity=2)
